#include "ASTSerialization.h"
#include "IRGenerator.h"
#include "SemanticAnalysis.h"
#include "SemanticException.h"

#include "llvm/IR/Verifier.h"
#include "llvm/Support/FileSystem.h"
#include "llvm/Support/Path.h"
#include "llvm/Support/raw_ostream.h"

#include <iostream>

void printUsage(int argc, char **argv) {
  std::cout << argv[0] << " <inputFile>" << std::endl;
}

int main(int argc, char **argv) {
  if (argc < 2) {
    printUsage(argc, argv);
    return 0;
  }

  std::string InputFile(argv[1]);
  lcpl::ASTDeserializer deserializer(InputFile);
  if (!deserializer.isValid()) {
    return -1;
  }

  auto program = deserializer.getRootNode();

  try {
    lcpl::SemanticAnalysis semanticAnalysis(program.get());
    semanticAnalysis.runAnalysis();

    std::string OutputFile = llvm::sys::path::filename(InputFile);
    lcpl::IRGenerator Generator(OutputFile, program.get(),
                                semanticAnalysis.getTypeTable(),
                                semanticAnalysis.getSymbolDefinitions());
    auto Module = Generator.runGenerator();
    if (!Module) {
      llvm::report_fatal_error("Couldn't generate module");
    }

    std::string filename(Module->getName().str() + ".ll");
    std::error_code err;
    llvm::raw_fd_ostream output(filename, err, llvm::sys::fs::F_Text);

    if (err) {
      llvm::report_fatal_error("Couldn't open output file: " + filename + "\n");
    }

    Module->print(output, nullptr);

    if (llvm::verifyModule(*Module, &llvm::errs())) {
      return -3;
    }
  } catch (lcpl::SemanticException &e) {
    llvm::errs() << e.what() << "\n";
    return -2;
  }

  return 0;
}
